//---------------------------------------------------------------------------
/**
 *  @file FileGenerator.cpp
 *
 *  $Id: FileGenerator.cpp,v 1.2 2001/05/18 11:07:01 chris Exp $
 *
 *  @author Christopher Kohlhoff <chris@kohlhoff.com>
 *
 *  This file contains a class for generating project and source files.
 */
//---------------------------------------------------------------------------
#include <vcl.h>
#pragma hdrstop
#include "FileGenerator.h"
#include <memory>
//---------------------------------------------------------------------------
TFileGenerator::TFileGenerator()
{
}
//---------------------------------------------------------------------------
TFileGenerator::~TFileGenerator()
{
}
//---------------------------------------------------------------------------
void TFileGenerator::SetValue(const String& key, const String& value)
{
  FValues[key] = value;
}
//---------------------------------------------------------------------------
String TFileGenerator::GetValue(const String& key) const
{
  std::map<String, String>::const_iterator i = FValues.find(key);
  if (i == FValues.end())
    return "";
  return i->second;
}
//---------------------------------------------------------------------------
String TFileGenerator::Generate(const String& templateName) const
{
  char module[MAX_PATH];
  GetModuleFileName(HInstance, module, MAX_PATH);
  String filename = ExtractFilePath(module) + templateName + ".txt";

  std::auto_ptr<TStrings> data(new TStringList);
  data->LoadFromFile(filename);
  ExpandConditionals(data.get());
  ExpandMacros(data.get());

  return data->Text;
}
//---------------------------------------------------------------------------
void TFileGenerator::ExpandConditionals(TStrings* strings) const
{
  int i = 0;
  while (i < strings->Count)
  {
    if (IsIfLine(strings->Strings[i]))
      i = ExpandLiveConditionalBranch(strings, i);
    else
      ++i;
  }
}
//---------------------------------------------------------------------------
int TFileGenerator::ExpandLiveConditionalBranch(TStrings* strings, int startRow) const
{
  if (ConditionalIsTrue(strings->Strings[startRow]))
  {
    strings->Delete(startRow);
    int i = startRow;

    // copy all lines until an 'else' or 'endif'
    while (i < strings->Count)
    {
      String line = strings->Strings[i];
      if (IsIfLine(line))
        i = ExpandLiveConditionalBranch(strings, i);
      else if (IsElseLine(line) || IsEndifLine(line))
        break;
      else
        ++i;
    }

    // delete the rest of the lines
    while (i < strings->Count)
    {
      String line = strings->Strings[i];
      if (IsIfLine(line))
        i = ExpandDeadConditionalBranch(strings, i);
      else
        strings->Delete(i);
      if (IsEndifLine(line))
        break;
    }

    return i;
  }
  else
  {
    strings->Delete(startRow);
    int i = startRow;

    // delete lines up to else or endif
    while (i < strings->Count)
    {
      String line = strings->Strings[i];
      if (IsIfLine(line))
        i = ExpandDeadConditionalBranch(strings, i);
      else if (IsEndifLine(line)) // leave endif for second loop to find
        break;
      else
      {
        strings->Delete(i);
        if (IsElseLine(line))
          break;
      }
    }

    // copy all lines until an 'endif'
    while (i < strings->Count)
    {
      String line = strings->Strings[i];
      if (IsIfLine(line))
        i = ExpandLiveConditionalBranch(strings, i);
      else if (IsEndifLine(line))
      {
        strings->Delete(i);
        break;
      }
      else
        ++i;
    }

    return i;
  }
}
//---------------------------------------------------------------------------
int TFileGenerator::ExpandDeadConditionalBranch(TStrings* strings, int startRow) const
{
  // delete opening 'if'
  strings->Delete(startRow);

  // delete subsequent rows until we find 'endif'
  int i = startRow;
  while (i < strings->Count)
  {
    String line = strings->Strings[i];
    if (IsIfLine(line))
      i = ExpandDeadConditionalBranch(strings, i);
    else
    {
      strings->Delete(i);
      if (IsEndifLine(line))
        break;
    }
  }

  return i;
}
//---------------------------------------------------------------------------
bool TFileGenerator::ConditionalIsTrue(String conditional) const
{
  conditional = ExpandMacros(conditional);

  // find the '==' in the condition
  int eqeq = conditional.Pos("==");
  if (!eqeq)
    return true;

  // extract the sides of the expression
  String first = conditional.SubString(5, eqeq - 5);
  String second = conditional.SubString(eqeq + 2, conditional.Length() - eqeq - 2);
  if (second[1] == '\'' && second[second.Length()] == '\'')
    second = second.SubString(2, second.Length() - 2);

  return (first == second);
}
//---------------------------------------------------------------------------
bool TFileGenerator::IsIfLine(String line) const
{
  return (line.SubString(1, 4) == "?if " && line[line.Length()] == '?');
}
//---------------------------------------------------------------------------
bool TFileGenerator::IsElseLine(String line) const
{
  return (line.SubString(1, 5) == "?else" && line[line.Length()] == '?');
}
//---------------------------------------------------------------------------
bool TFileGenerator::IsEndifLine(String line) const
{
  return (line.SubString(1, 6) == "?endif" && line[line.Length()] == '?');
}
//---------------------------------------------------------------------------
void TFileGenerator::ExpandMacros(TStrings* strings) const
{
  for (int i = 0; i < strings->Count; ++i)
    strings->Strings[i] = ExpandMacros(strings->Strings[i]);
}
//---------------------------------------------------------------------------
String TFileGenerator::ExpandMacros(String line) const
{
  String newLine;
  String macroName;
  for (int i = 1; i <= line.Length(); ++i)
  {
    if (line[i] == '@')
    {
      int nextAt;
      for (nextAt = i + 1; nextAt <= line.Length(); ++nextAt)
        if (line[nextAt] == '@')
          break;
      if (nextAt <= line.Length())
        newLine += GetValue(line.SubString(i + 1, nextAt - i - 1));
      else
        newLine += line.SubString(i, nextAt - i);
      i = nextAt;
    }
    else
      newLine += line[i];
  }

  return newLine;
}
//---------------------------------------------------------------------------
